#include "SDL_error.h"
#include "SDL_video.h"
#include <SDL2/SDL.h>
#include <SDL2/SDL_events.h>
#include "glad/glad.h"
#include <cstdint>
#include <cstdio>
#include <stdbool.h>
#include <string>
#include <vector>

#define WINDOW_WIDTH 1280
#define WINDOW_HEIGHT 720

enum exit_codes : int {
  EXIT_CODE_SUCCESS,
  EXIT_CODE_SDL_INIT_FAILED,
  EXIT_CODE_WINDOW_CREATION_FAILED,
  EXIT_CODE_OPENGL_CONTEXT_FAILED,
  EXIT_CODE_GLAD_LOADER_FAILED,
};

struct App {
  SDL_Window    *window;
  SDL_GLContext context;
  SDL_Event     event;

  // Defines the different attributes in the vertex data and how to access them.
  // Think of it as a specification or a C struct that defines the types of data stored for each
  // vertex.
  GLuint        vao;

  // The buffer that contains all the vertex data
  // e.g. assume 3 vertices with position, colour and uv, the buffer would look like this
  //  -----------------------------------------------------------------------------------
  // | position1 | colour1 | uv1 | position2 | colour2 | uv2 | position3 | colour3 | uv3 |
  //  -----------------------------------------------------------------------------------
  GLuint        vbo;

  GLuint        shader_program;
  bool          running;
};

int    init                    (App &app);
void   create_vertex_spec      (App &app);
void   create_graphics_pipeline(App &app);
void   run_main_loop           (App &app);
void   cleanup                 (App &app);
GLuint create_shader_program   (const std::string &vertex_shader_source, const std::string &fragment_shader_source);
GLuint compile_shader          (GLuint type, const std::string &source);

int main() {
  App app    = {};
  int result = init(app);
  if (result != EXIT_CODE_SUCCESS) {
    return result;
  }

  // Setup the geometry
  create_vertex_spec(app);

  // Setup graphics pipeline. At the very minimum creating vertex and fragment shaders
  create_graphics_pipeline(app);

  run_main_loop(app);

  cleanup(app);

  return EXIT_CODE_SUCCESS;
}

int init(App &app) {
  if (SDL_Init(SDL_INIT_EVERYTHING) != 0) {
    fprintf(stderr, "Failed to initialise SDL: %s", SDL_GetError());
    return EXIT_CODE_SDL_INIT_FAILED;
  }

  uint32_t flags = SDL_WINDOW_SHOWN | SDL_WINDOW_OPENGL | SDL_WINDOW_ALLOW_HIGHDPI;
  app.window     = SDL_CreateWindow("Window", SDL_WINDOWPOS_CENTERED, SDL_WINDOWPOS_CENTERED,
                                    WINDOW_WIDTH, WINDOW_HEIGHT, flags);
  if (!app.window) {
    fprintf(stderr, "Failed to create window: %s", SDL_GetError());
    return EXIT_CODE_WINDOW_CREATION_FAILED;
  }

  SDL_GL_SetAttribute(SDL_GL_CONTEXT_MAJOR_VERSION, 4);
  SDL_GL_SetAttribute(SDL_GL_CONTEXT_MINOR_VERSION, 6);
  SDL_GL_SetAttribute(SDL_GL_CONTEXT_PROFILE_MASK, SDL_GL_CONTEXT_PROFILE_CORE);
  SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
  SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, 24);

  app.context = SDL_GL_CreateContext(app.window);
  if (!app.context) {
    fprintf(stderr, "Failed to create OpenGL context: %s", SDL_GetError());
    return EXIT_CODE_OPENGL_CONTEXT_FAILED;
  }

  // Init GLAD
  if (!gladLoadGLLoader(SDL_GL_GetProcAddress)) {
    fprintf(stderr, "Failed to initialise glad");
    return EXIT_CODE_GLAD_LOADER_FAILED;
  }

  printf("%s\n%s\n%s\n%s\n",
         glGetString(GL_VENDOR),
         glGetString(GL_RENDERER),
         glGetString(GL_VERSION),
         glGetString(GL_SHADING_LANGUAGE_VERSION)
  );

  return EXIT_CODE_SUCCESS;
}

void create_vertex_spec(App &app) {
  const std::vector<GLfloat> vertices = {
    -0.8f, -0.8f, 0.0f, // position
     0.8f, -0.8f, 0.0f, // position
     0.0f,  0.8f, 0.0f, // position
  };

  // Create and activate the VAO
  glGenVertexArrays(1, &app.vao);
  glBindVertexArray(app.vao);

  glGenBuffers(1, &app.vbo);
  glBindBuffer(GL_ARRAY_BUFFER, app.vbo);
  glBufferData(GL_ARRAY_BUFFER,
               vertices.size() * sizeof(GLfloat),
               (void *)vertices.data(),
               GL_STATIC_DRAW
  );

  // Defines the vertex attributes and their indices. This defines the attribute for the currently
  // bound VAO
  glEnableVertexAttribArray(0);

  // Defines the number of components for the attribute, its type, the stride between the component
  // for each vertex and the offset of the first instance of the component in the buffer
  glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 0, (void *)0);

  // Cleanup.
  // The order matters. Unbind VAO first then disable the attributes. If you do it the
  // other way around, the VAO will store that the attributes are disabled and nothing will be
  // drawn during the rendering stage
  glBindVertexArray(0);
  glBindBuffer(GL_ARRAY_BUFFER, 0);
  glDisableVertexAttribArray(0);
}

void create_graphics_pipeline(App &app) {
  const std::string vs_source =
    "#version 460 core\n"
    "in vec4 position;\n"
    "void main() {\n"
    "  gl_Position = vec4(position.x, position.y, position.z, position.w);\n"
    "}";
  const std::string fs_source =
    "#version 460 core\n"
    "out vec4 color;\n"
    "void main() {\n"
    "  color = vec4(0.93f, 0.42f, 0.35f, 1.0f);\n"
    "}";

  app.shader_program = create_shader_program(vs_source, fs_source);
}

void run_main_loop(App &app) {
  app.running = true;

  while (app.running) {
    while (SDL_PollEvent(&app.event)) {
      switch (app.event.type) {
      case SDL_QUIT:
        app.running = false;
        break;
      }
    }

    // Pre draw setup
    glDisable(GL_DEPTH_TEST);
    glDisable(GL_CULL_FACE);

    glViewport  (0, 0, WINDOW_WIDTH, WINDOW_HEIGHT);
    glClearColor(0.36f, 0.34f, 0.42f, 1.0f);
    glClear(GL_DEPTH_BUFFER_BIT | GL_COLOR_BUFFER_BIT);

    glUseProgram(app.shader_program);
    // End pre draw setup

    // Draw call
    glBindVertexArray(app.vao);
    glBindBuffer(GL_ARRAY_BUFFER, app.vbo); // Is this actually needed?
    glDrawArrays(GL_TRIANGLES, 0, 3);
    // End draw call

    // Not necessary if we only have one shader program
    glUseProgram(0);

    SDL_GL_SwapWindow(app.window);
  }
}

void cleanup(App &app) {
  SDL_GL_DeleteContext(app.context);
  SDL_DestroyWindow(app.window);
  SDL_Quit();
}

GLuint create_shader_program(const std::string &vertex_shader_source, const std::string &fragment_shader_source) {
  GLuint shader_program  = glCreateProgram();
  GLuint vertex_shader   = compile_shader(GL_VERTEX_SHADER, vertex_shader_source);
  GLuint fragment_shader = compile_shader(GL_FRAGMENT_SHADER, fragment_shader_source);

  glAttachShader(shader_program, vertex_shader);
  glAttachShader(shader_program, fragment_shader);
  glLinkProgram (shader_program);

  GLint link_status;
  glGetProgramiv(shader_program, GL_LINK_STATUS, &link_status);
  if (link_status != GL_TRUE) {
    GLsizei log_length = 0;
    GLchar message[1024];
    glGetProgramInfoLog(shader_program, 1024, &log_length, message);
    fprintf(stderr, "Failed to link program: %s\n", message);
  }

  // Validate the shader_program
  glValidateProgram(shader_program);
  GLint validation_status;
  glGetProgramiv(shader_program, GL_VALIDATE_STATUS, &validation_status);
  if (validation_status != GL_TRUE) {
    GLsizei log_length = 0;
    GLchar message[1024];
    glGetProgramInfoLog(shader_program, 1024, &log_length, message);
    fprintf(stderr, "Failed to validate program: %s\n", message);
  }

  return shader_program;
}

GLuint compile_shader(GLuint type, const std::string &source) {
  GLuint shader             = glCreateShader(type);
  const char *shader_source = source.c_str();

  glShaderSource (shader, 1, &shader_source, NULL);
  glCompileShader(shader);

  GLint compile_status;
  glGetShaderiv(shader, GL_COMPILE_STATUS, &compile_status);
  if (compile_status != GL_TRUE) {
    GLsizei log_length = 0;
    GLchar message[1024];
    glGetShaderInfoLog(shader, 1024, &log_length, message);
    fprintf(stderr, "Failed to compile shader: %s\n", message);
  }

  return shader;
}
